@@ -1,6 +1,7 @@
 /*
   This is the machine-specific part for Intel 386 compatible processors
 
+  Authors: Anton Ertl, Bernd Paysan, Jens Wilke
   Copyright (C) 1995,1996,1997,1998,2000,2003,2004,2005,2006,2007,2008,2012,2013,2014,2016,2018 Free Software Foundation, Inc.
 
   This file is part of Gforth.
@@ -86,11 +87,7 @@
      /* gcc 2.95 has a better register allocater */
 #    define SPREG asm("%esi")
 #    define RPREG asm("%edi")
-#    ifdef NO_IP
-#     define spbREG asm("%ebx")
-#    else
-#     define IPREG asm("%ebx")
-#    endif
+#    define IPREG asm("%ebx")
 /* ebp leads to broken code (gcc-3.0); eax, ecx, edx produce compile errors */
 #    define TOSREG asm("%ecx")
 /* ecx works only for TOS, and eax, edx don't work for anything (gcc-3.0) */
@@ -127,11 +124,7 @@
 #   if ((__GNUC__==2 && defined(__GNUC_MINOR__) && __GNUC_MINOR__>=95) || (__GNUC__>2))
 #    define SPREG asm("%esi")
 #    define RPREG asm("%edi")
-#    ifdef NO_IP
-#     define spbREG asm("%ebx")
-#    else
-#     define IPREG asm("%ebx")
-#    endif
+#    define IPREG asm("%ebx")
 #   else /* !(gcc-2.95 or later) */
 #    define SPREG asm("%ebx")
 #   endif  /* !(gcc-2.95 or later) */
