@@ -1,5 +1,6 @@
 /* Gforth virtual machine (aka inner interpreter)
 
+  Authors: Anton Ertl, Bernd Paysan, David Kühling
   Copyright (C) 1995,1996,1997,1998,2000,2003,2004,2005,2006,2007,2008,2010,2011,2012,2013,2014,2015,2016 Free Software Foundation, Inc.
 
   This file is part of Gforth.
@@ -26,6 +27,7 @@
 
 #include "config.h"
 #include "forth.h"
+#include "symver.h"
 #include <ctype.h>
 #include <stdio.h>
 #include <string.h>
@@ -96,11 +98,9 @@ extern int debug;
 extern int gforth_memcmp(const char * s1, const char * s2, size_t n);
 extern Char *gforth_memmove(Char * dest, const Char* src, Cell n);
 extern Char *gforth_memset(Char * s, Cell c, UCell n);
-extern Char *gforth_memcpy(Char * dest, const Char* src, Cell n);
 #define memcmp(s1,s2,n) gforth_memcmp(s1,s2,n)
 #define memmove(a,b,c) gforth_memmove(a,b,c)
 #define memset(a,b,c) gforth_memset(a,b,c)
-#define memcpy(a,b,c) gforth_memcpy(a,b,c)
 #endif
 
 #define NEWLINE	'\n'
@@ -155,11 +155,7 @@ extern Char *gforth_memcpy(Char * dest, const Char* src, Cell n);
 
 #define vm_Cell2Cell(_x,_y)		(_y=_x)
 
-#ifdef NO_IP
-#define IMM_ARG(access,value)		(VARIANT(value))
-#else
 #define IMM_ARG(access,value)		(access)
-#endif
 
 /* if machine.h has not defined explicit registers, define them as implicit */
 #ifndef IPREG
@@ -296,8 +292,26 @@ void throw(int code)
      I_##name:
 #define LABEL_UU(name) H_##name: MAYBE_UNUSED asm(ASMCOMMENT "I " #name); \
      I_##name: MAYBE_UNUSED
-#define LABEL3(name) J_##name: asm(ASMCOMMENT "J " #name);
-#define LABEL3_UU(name) J_##name: MAYBE_UNUSED asm(ASMCOMMENT "J " #name);
+#define LABEL3(name) J_##name: { \
+    asm(ASMCOMMENT "J " #name);  \
+    asm(ASMCOMMENT "J " #name);  \
+    asm(ASMCOMMENT "J " #name);  \
+    asm(ASMCOMMENT "J " #name);  \
+    asm(ASMCOMMENT "J " #name);  \
+    asm(ASMCOMMENT "J " #name);  \
+    asm(ASMCOMMENT "J " #name);  \
+    asm(ASMCOMMENT "J " #name);  \
+  }
+#define LABEL3_UU(name) J_##name: MAYBE_UNUSED { \
+    asm(ASMCOMMENT "J " #name);                  \
+    asm(ASMCOMMENT "J " #name);                  \
+    asm(ASMCOMMENT "J " #name);                  \
+    asm(ASMCOMMENT "J " #name);                  \
+    asm(ASMCOMMENT "J " #name);  \
+    asm(ASMCOMMENT "J " #name);  \
+    asm(ASMCOMMENT "J " #name);  \
+    asm(ASMCOMMENT "J " #name);  \
+  }
 
 #elif ENGINE==2
 /* variant with padding between VM instructions for finding out
@@ -313,8 +327,26 @@ void throw(int code)
    some constants after the final branch, and may refer to them from
    the code before label3.  Since we don't copy the constants, we have
    to make sure that such code is recognized as non-relocatable. */
-#define LABEL3(name) J_##name: asm(ASMCOMMENT "J " #name); SKIP16;
-#define LABEL3_UU(name) J_##name: MAYBE_UNUSED asm(ASMCOMMENT "J " #name); SKIP16;
+#define LABEL3(name) J_##name: { \
+    asm(ASMCOMMENT "J " #name);  \
+    asm(ASMCOMMENT "J " #name);  \
+    asm(ASMCOMMENT "J " #name);  \
+    asm(ASMCOMMENT "J " #name);  \
+    asm(ASMCOMMENT "J " #name);  \
+    asm(ASMCOMMENT "J " #name);  \
+    asm(ASMCOMMENT "J " #name);  \
+    asm(ASMCOMMENT "J " #name);  \
+  } SKIP16;
+#define LABEL3_UU(name) J_##name: MAYBE_UNUSED { \
+    asm(ASMCOMMENT "J " #name);                  \
+    asm(ASMCOMMENT "J " #name);                  \
+    asm(ASMCOMMENT "J " #name);                  \
+    asm(ASMCOMMENT "J " #name);                  \
+    asm(ASMCOMMENT "J " #name);  \
+    asm(ASMCOMMENT "J " #name);  \
+    asm(ASMCOMMENT "J " #name);  \
+    asm(ASMCOMMENT "J " #name);  \
+  } SKIP16;
 
 #elif ENGINE==3
 /* variant with different immediate arguments for finding out
@@ -326,8 +358,26 @@ void throw(int code)
     I_##name:
 #define LABEL_UU(name) H_##name: MAYBE_UNUSED asm(ASMCOMMENT "I " #name); \
     I_##name: MAYBE_UNUSED 
-#define LABEL3(name) J_##name: asm(ASMCOMMENT "J " #name);
-#define LABEL3_UU(name) J_##name: MAYBE_UNUSED asm(ASMCOMMENT "J " #name);
+#define LABEL3(name) J_##name: { \
+    asm(ASMCOMMENT "J " #name);  \
+    asm(ASMCOMMENT "J " #name);  \
+    asm(ASMCOMMENT "J " #name);  \
+    asm(ASMCOMMENT "J " #name);  \
+    asm(ASMCOMMENT "J " #name);  \
+    asm(ASMCOMMENT "J " #name);  \
+    asm(ASMCOMMENT "J " #name);  \
+    asm(ASMCOMMENT "J " #name);  \
+  }
+#define LABEL3_UU(name) J_##name: MAYBE_UNUSED { \
+    asm(ASMCOMMENT "J " #name);                  \
+    asm(ASMCOMMENT "J " #name);                  \
+    asm(ASMCOMMENT "J " #name);                  \
+    asm(ASMCOMMENT "J " #name);                  \
+    asm(ASMCOMMENT "J " #name);  \
+    asm(ASMCOMMENT "J " #name);  \
+    asm(ASMCOMMENT "J " #name);  \
+    asm(ASMCOMMENT "J " #name);  \
+  }
 #else
 #error illegal ENGINE value
 #endif /* ENGINE */
@@ -359,9 +409,7 @@ Label *gforth_engine(Xt *ip0 sr_proto)
 # undef saved_ip
   Xt* MAYBE_UNUSED saved_ip;
 #endif /* !defined(GFORTH_DEBUGGING) */
-#ifndef NO_IP
   register Xt *ip IPREG = ip0;
-#endif
   register Cell *sp SPREG = gforth_SP;
   register Float *fp FPREG = gforth_FP;
   register Address lp LPREG = gforth_LP;
@@ -488,13 +536,9 @@ Label *gforth_engine(Xt *ip0 sr_proto)
 
   IF_fpTOS(fpTOS = fp[0]);
 /*  prep_terminal(); */
-#ifdef NO_IP
-  goto *(*(Label *)ip0);
-#else
   SET_IP(ip);
   SUPER_END; /* count the first block, too */
   NEXT;
-#endif
 
 #ifdef CPU_DEP3
   CPU_DEP3
