@@ -1,5 +1,6 @@
 /* common header file
 
+  Authors: Bernd Paysan, Anton Ertl, David Kühling, Jens Wilke, Neal Crook
   Copyright (C) 1995,1996,1997,1998,2000,2003,2004,2005,2006,2007,2008,2009,2010,2011,2012,2013,2014,2015,2016,2017,2018 Free Software Foundation, Inc.
 
   This file is part of Gforth.
@@ -114,12 +115,10 @@ extern void* (*realloc_l)(void* addr, size_t size);
 #define DODEFER	4
 #define DOFIELD	5
 #define DOVAL	6
-#define DODOES	7
+#define DODOES  7
 #define DOABICODE	8
 #define DOSEMIABICODE   9
-#define DOEXTRA	10
-#define DODOESXT 11
-#define DOER_MAX        11
+#define DOER_MAX 9
 
 #include "machine.h"
 
@@ -143,6 +142,8 @@ typedef unsigned OCTABYTE_TYPE UOctabyte;
 #define LH(x)		((x)&HALFCELL_MASK)
 #define L2U(x)		(((UCell)(x))<<HALFCELL_BITS)
 #define HIGHBIT(x)	(((UCell)(x))>>(CELL_BITS-1))
+#define SECTION(x)      (((UCell)(x))>>(CELL_BITS-8))
+#define INSECTION(x)    (((UCell)(x))&((~(UCell)0)>>8))
 
 #define F_TRUE (FLAG(0==0))
 #define F_FALSE (FLAG(0!=0))
@@ -303,7 +304,7 @@ typedef union {
 typedef Label *Xt;
 
 /* PFA gives the parameter field address corresponding to a cfa */
-#define PFA(cfa)	(((Cell *)cfa)+2)
+#define PFA(cfa)	(((Cell *)cfa)+1)
 /* PFA1 is a special version for use just after a NEXT1 */
 #define PFA1(cfa)	PFA(cfa)
 /* CODE_ADDRESS is the address of the code jumped to through the code field */
@@ -316,17 +317,21 @@ typedef Label *Xt;
 #  define DOES_CA ((Label)&xts[DODOES])
 #endif /* defined(DOUBLY_INDIRECT) */
 
-#define DOES_CODE1(cfa)	 ((Xt *)(((Cell *)cfa)[1]))
-#define DOES_CODEXT(cfa) ((Xt)(((Cell *)cfa)[1]))
+/* Extra is used for DOES */
+#define VTLINK 0
+#define VTCOMPILE 1
+#define VTTO 2
+#define VTDEFER 3
+#define VTEXTRA 4
+#define VT2INT 5
+#define VT2COMP 6
+#define VT2STRING 7
+#define VT2LINK 8
+#define EXTRA_CODE(cfa) ((Xt *)(((Cell **)cfa)[-1][VTEXTRA]))
+#define EXTRA_CODEXT(cfa) ((Xt)(((Cell **)cfa)[-1][VTEXTRA]))
 
 /* MAKE_CF creates an appropriate code field at the cfa;
    ca is the code address */
-#define VTLINK 0
-#define VTCOMPILE 1
-#define VTLIT 2
-#define VTEXTRA 3
-#define VTTO 4
-#define EXTRA_CODE(cfa) ((Xt *)(((Cell **)cfa)[-1][VTEXTRA]))
 #define MAKE_CF(cfa,ca) ((*(Label *)(cfa)) = ((Label)ca))
 /* make a code field for a defining-word-defined word */
 
@@ -379,11 +384,10 @@ typedef struct {
 
 typedef struct {
   Address base;		/* base address of image (0 if relocatable) */
-  UCell checksum;	/* checksum of ca's to protect against some
-			   incompatible	binary/executable combinations
-			   (0 if relocatable) */
-  UCell image_size;	/* all sizes in bytes */
   UCell dict_size;
+  Address image_dp;	/* all sizes in bytes */
+  Address sect_name;
+  Address sect_locs;
   UCell data_stack_size;
   UCell fp_stack_size;
   UCell return_stack_size;
@@ -393,11 +397,20 @@ typedef struct {
   Xt *quit_entry;
   Xt *execute_entry;
   Xt *find_entry;
+  UCell checksum;	/* checksum of ca's to protect against some
+			   incompatible	binary/executable combinations
+			   (0 if relocatable) */
   Label *xt_base;         /* base of DOUBLE_INDIRECT xts[], for comp-i.fs */
   Label *label_base;      /* base of DOUBLE_INDIRECT labels[], for comp-i.fs */
 } ImageHeader;
 /* the image-header is created in main.fs */
 
+typedef struct {
+  Address base;
+  Address end;
+  Address dp;
+} SectionHeader;
+
 #ifdef HAS_F83HEADERSTRING
 struct F83Name {
   struct F83Name *next;  /* the link field for old hands */
@@ -559,6 +572,7 @@ void vm_count_block(Xt *ip);
 /* dynamic superinstruction stuff */
 void compile_prim1(Cell *start);
 void finish_code(void);
+void finish_code_barrier(void);
 int forget_dyncode(Address code);
 Label decompile_code(Label prim);
 
@@ -598,10 +612,6 @@ extern Cell const * gforth_pointers(Cell n);
 extern void gforth_callback(Xt* fcall, void * alist);
 #endif
 
-#ifdef NO_IP
-extern Label next_code;
-#endif
-
 #ifdef HAS_FILE
 extern char* fileattr[6];
 extern char* pfileattr[6];
