@@ -1,6 +1,7 @@
 /* command line interpretation, image loading etc. for Gforth
 
 
+  Authors: Anton Ertl, Bernd Paysan, Jens Wilke, David Kühling
   Copyright (C) 1995,1996,1997,1998,2000,2003,2004,2005,2006,2007,2008,2009,2010,2011,2012,2013,2014,2015,2016,2017,2018 Free Software Foundation, Inc.
 
   This file is part of Gforth.
@@ -21,6 +22,7 @@
 
 #include "config.h"
 #include "forth.h"
+#include "symver.h"
 #include <errno.h>
 #include <ctype.h>
 #include <stdio.h>
@@ -97,10 +99,6 @@ void gforth_callback(Xt* fcall, void * alist)
 }
 #endif
 
-#ifdef NO_IP
-Label next_code;
-#endif
-
 #ifdef HAS_FILE
 char* fileattr[6]={"rb","rb","r+b","r+b","wb","wb"};
 char* pfileattr[6]={"r","r","r+","r+","w","w"};
@@ -259,11 +257,6 @@ Char *gforth_memset(Char * s, Cell c, UCell n)
 {
   return memset(s, c, n);
 }
-
-Char *gforth_memcpy(Char * dest, const Char* src, Cell n)
-{
-  return memcpy(dest, src, n);
-}
 #endif
 
 static Cell max(Cell a, Cell b)
@@ -293,13 +286,16 @@ static Cell min(Cell a, Cell b)
  *  data (size in ImageHeader.image_size)
  *  tags ((if relocatable, 1 bit/data cell)
  *
+ * If the image has sections, they follow after the main image with each
+ * section starting with the magic "Section.". A section starts with the
+ * section header (see section.fs and forth.h), has data and tags.
+ *
  * tag==1 means that the corresponding word is an address;
  * If the word is >=0, the address is within the image;
- * addresses within the image are given relative to the start of the image.
+ * addresses within the image are given relative to the start of the section.
+ * bits MSB..MSB-7 (8 bits) index the section
  * If the word =-1 (CF_NIL), the address is NIL,
- * If the word is <CF_NIL and >CF(DODOES), it's a CFA (:, Create, ...)
- * If the word =CF(DODOES), it's a DOES> CFA
- * !! ABI-CODE and ;ABI-CODE
+ * If the word is <CF_NIL and >=CF(DOER_MAX), it's a CFA (:, Create, ...)
  * If the word is <CF(DOER_MAX) and bit 14 is set, it's the xt of a primitive
  * If the word is <CF(DOER_MAX) and bit 14 is clear, 
  *                                        it's the threaded code of a primitive
@@ -322,7 +318,7 @@ Cell groups[32] = {
 };
 
 static unsigned char *branch_targets(Cell *image, const unsigned char *bitstring,
-			      int size, Cell base)
+				     int size, Cell base, int sect)
      /* produce a bitmask marking all the branch targets */
 {
   int i=0, j, k, steps=(((size-1)/sizeof(Cell))/RELINFOBITS)+1;
@@ -336,8 +332,9 @@ static unsigned char *branch_targets(Cell *image, const unsigned char *bitstring
       if(bits & (1U << (RELINFOBITS-1))) {
 	assert(i*sizeof(Cell) < size);
         token=image[i];
-	if (token>=base) { /* relocatable address */
-	  UCell bitnum=(token-base)/sizeof(Cell);
+	if ((token>=base) &&
+	    (SECTION(token) == sect)) { /* relocatable address */
+	  UCell bitnum=(INSECTION(token)-INSECTION(base))/sizeof(Cell);
 	  if (bitnum/RELINFOBITS < (UCell)steps)
 	    result[bitnum/RELINFOBITS] |= 1U << ((~bitnum)&(RELINFOBITS-1));
 	}
@@ -347,10 +344,10 @@ static unsigned char *branch_targets(Cell *image, const unsigned char *bitstring
   return result;
 }
 
-void gforth_relocate(Cell *image, const Char *bitstring, 
-		     UCell size, Cell base, Label symbols[])
+void gforth_relocate(Address sections[], Char *bitstrings[], 
+		     UCell sizes[], Cell bases[], Label symbols[])
 {
-  int i=0, j, k, steps=(((size-1)/sizeof(Cell))/RELINFOBITS)+1;
+  int i=0, j, k;
   Cell token;
   char bits;
   Cell max_symbols;
@@ -358,109 +355,126 @@ void gforth_relocate(Cell *image, const Char *bitstring,
    * A virtual start address that's the real start address minus 
    * the one in the image 
    */
-  Cell *start = (Cell * ) (((void *) image) - ((void *) base));
-  unsigned char *targets = branch_targets(image, bitstring, size, base);
+  int ii;
+  for (ii=0; ii<0x100; ii++) {
+    Char * bitstring=bitstrings[ii];
+    Cell * image=(Cell*)sections[ii];
+    UCell size=sizes[ii];
+    Cell base=bases[ii];
 
-  /* group index into table */
-  if(groups[31]==0) {
-    int groupsum=0;
-    for(i=0; i<32; i++) {
-      groupsum += groups[i];
-      groups[i] = groupsum;
-      /* printf("group[%d]=%d\n",i,groupsum); */
+    int steps=(((size-1)/sizeof(Cell))/RELINFOBITS)+1;
+
+    debugp(stderr, "relocate section %i, %p:%p\n", ii, base, size);
+    
+    if(!bitstring) break;
+    
+    unsigned char *targets = branch_targets(image, bitstring, size, base, ii);
+    
+    /* group index into table */
+    if(groups[31]==0) {
+      int groupsum=0;
+      for(i=0; i<32; i++) {
+	groupsum += groups[i];
+	groups[i] = groupsum;
+	/* printf("group[%d]=%d\n",i,groupsum); */
+      }
+      i=0;
     }
-    i=0;
-  }
-  
-/* printf("relocating to %x[%x] start=%x base=%x\n", image, size, start, base); */
-  
-  for (max_symbols=0; symbols[max_symbols]!=0; max_symbols++)
-    ;
-  max_symbols--;
-
-  for(k=0; k<steps; k++) {
-    for(j=0, bits=bitstring[k]; j<RELINFOBITS; j++, i++, bits<<=1) {
-      /*      fprintf(stderr,"relocate: image[%d]\n", i);*/
-      if(bits & (1U << (RELINFOBITS-1))) {
-	assert(i*sizeof(Cell) < size);
-	/* fprintf(stderr,"relocate: image[%d]=%d of %d\n", i, image[i], size/sizeof(Cell)); */
-        token=image[i];
-	if(token<0) {
-	  int group = (-token & 0x3E00) >> 9;
-	  if(group == 0) {
-	    switch(token|0x4000) {
-	    case CF_NIL      : image[i]=0; break;
+    
+    /* printf("relocating to %x[%x] start=%x base=%x\n", image, size, start, base); */
+    
+    for (max_symbols=0; symbols[max_symbols]!=0; max_symbols++)
+      ;
+    max_symbols--;
+    
+    for(i=k=0; k<steps; k++) {
+      for(j=0, bits=bitstring[k]; j<RELINFOBITS; j++, i++, bits<<=1) {
+	/*      fprintf(stderr,"relocate: image[%d]\n", i);*/
+	if(bits & (1U << (RELINFOBITS-1))) {
+	  // debugp(stderr,"relocate: image[%d]=%d of %d\n", i, image[i], size/sizeof(Cell));
+	  assert(i*sizeof(Cell) < size);
+	  token=image[i];
+	  if(SECTION(token)==0xFF) {
+	    int group = (-token & 0x3E00) >> 9;
+	    if(group == 0) {
+	      switch(token|0x4000) {
+	      case CF_NIL      : image[i]=0; break;
 #if !defined(DOUBLY_INDIRECT)
-	    case CF(DOCOL)   :
-	    case CF(DOVAR)   :
-	    case CF(DOCON)   :
-	    case CF(DOVAL)   :
-	    case CF(DOUSER)  : 
-	    case CF(DODEFER) : 
-	    case CF(DOFIELD) : 
-	    case CF(DODOES)  :
-	    case CF(DOABICODE) :
-	    case CF(DOSEMIABICODE): 
-	    case CF(DOEXTRA): 
-	    case CF(DODOESXT): 
-	      MAKE_CF(image+i,symbols[CF(token)]); break;
+	      case CF(DOER_MAX) ... CF(DOCOL):
+		MAKE_CF(image+i,symbols[CF(token)]); break;
 #endif /* !defined(DOUBLY_INDIRECT) */
-	    default          : /* backward compatibility */
-/*	      printf("Code field generation image[%x]:=CFA(%x)\n",
-		     i, CF(image[i])); */
-	      if (CF((token | 0x4000))<max_symbols) {
-		image[i]=(Cell)CFA(CF(token));
+	      default          : /* backward compatibility */
+		/*	      printf("Code field generation image[%x]:=CFA(%x)\n",
+			      i, CF(image[i])); */
+		if (CF((token | 0x4000))<max_symbols) {
+		  image[i]=(Cell)CFA(CF(token));
+#ifdef DIRECT_THREADED
+		  if ((token & 0x4000) == 0) { /* threaded code, no CFA */
+		    if (targets[k] & (1U<<(RELINFOBITS-1-j)))
+		      compile_prim1(0);
+		    compile_prim1(&image[i]);
+		  }
+#endif
+		} else {
+		  Char * dumpa = (Char*)&image[i];
+		  for(; dumpa < (Char*)&image[i+8]; dumpa++) {
+		    fprintf(stderr, "%02x ", *dumpa);
+		  }
+		  fprintf(stderr, "\n");
+		  fprintf(stderr,"Primitive %ld used in this image at %p (offset $%x) is not implemented by this\n engine (%s); executing this code will crash.\n",(long)CF(token), &image[i], i, PACKAGE_VERSION);
+		}
+	      }
+	    } else {
+	      int tok = -token & 0x1FF;
+	      if (tok < (groups[group+1]-groups[group])) {
+#if defined(DOUBLY_INDIRECT)
+		image[i]=(Cell)CFA(((groups[group]+tok) | (CF(token) & 0x4000)));
+#else
+		image[i]=(Cell)CFA((groups[group]+tok));
+#endif
 #ifdef DIRECT_THREADED
 		if ((token & 0x4000) == 0) { /* threaded code, no CFA */
 		  if (targets[k] & (1U<<(RELINFOBITS-1-j)))
 		    compile_prim1(0);
 		  compile_prim1(&image[i]);
+		} else if((token & 0x8000) == 0) { /* special CFA */
+		  /* debugp(stderr, "image[%x] = symbols[%x]\n", i, groups[group]+tok); */
+		  MAKE_CF(image+i,symbols[groups[group]+tok]);
 		}
 #endif
-	      } else
-		fprintf(stderr,"Primitive %ld used in this image at %p (offset $%x) is not implemented by this\n engine (%s); executing this code will crash.\n",(long)CF(token), &image[i], i, PACKAGE_VERSION);
+#if defined(DOUBLY_INDIRECT) || defined(INDIRECT_THREADED)
+		if((token & 0x8000) == 0) { /* special CFA */
+		  /* debugp(stderr, "image[%x] = symbols[%x] = %p\n", i, groups[group]+tok, symbols[groups[group]+tok]); */
+		  MAKE_CF(image+i,symbols[groups[group]+tok]);
+		}
+#endif
+	      } else {
+		Char * dumpa = (Char*)&image[i];
+		for(; dumpa < (Char*)&image[i+8]; dumpa++) {
+		  fprintf(stderr, "%02x ", *dumpa);
+		}
+		fprintf(stderr, "\n");
+		fprintf(stderr,"Primitive %lx, %d of group %d used in this image at %p (offset $%x) is not implemented by this\n engine (%s); executing this code will crash.\n", (long)-token, tok, group, &image[i],i,PACKAGE_VERSION);
+	      }
 	    }
 	  } else {
-	    int tok = -token & 0x1FF;
-	    if (tok < (groups[group+1]-groups[group])) {
-#if defined(DOUBLY_INDIRECT)
-	      image[i]=(Cell)CFA(((groups[group]+tok) | (CF(token) & 0x4000)));
-#else
-	      image[i]=(Cell)CFA((groups[group]+tok));
-#endif
-#ifdef DIRECT_THREADED
-	      if ((token & 0x4000) == 0) { /* threaded code, no CFA */
-		if (targets[k] & (1U<<(RELINFOBITS-1-j)))
-		  compile_prim1(0);
-		compile_prim1(&image[i]);
-	      } else if((token & 0x8000) == 0) { /* special CFA */
-		/* debugp(stderr, "image[%x] = symbols[%x]\n", i, groups[group]+tok); */
-		MAKE_CF(image+i,symbols[groups[group]+tok]);
-	      }
-#endif
-#if defined(DOUBLY_INDIRECT) || defined(INDIRECT_THREADED)
-	      if((token & 0x8000) == 0) { /* special CFA */
-		/* debugp(stderr, "image[%x] = symbols[%x] = %p\n", i, groups[group]+tok, symbols[groups[group]+tok]); */
-		MAKE_CF(image+i,symbols[groups[group]+tok]);
-	      }
-#endif
-	    } else
-	      fprintf(stderr,"Primitive %lx, %d of group %d used in this image at %p (offset $%x) is not implemented by this\n engine (%s); executing this code will crash.\n", (long)-token, tok, group, &image[i],i,PACKAGE_VERSION);
+	    /* if base is > 0: 0 is a null reference so don't adjust*/
+	    if (token>=base) {
+	      UCell sec = SECTION(token);
+	      UCell start = (Cell) (((void *) sections[sec]) - ((void *) bases[sec]));
+	      image[i]=start+INSECTION(token);
+	    } else if(token!=0) {
+	      fprintf(stderr, "tagged item image[%x]=%llx unrelocated\n", i, (long long)image[i]);
+	    }
 	  }
-	} else {
-          /* if base is > 0: 0 is a null reference so don't adjust*/
-          if (token>=base) {
-            image[i]+=(Cell)start;
-          } else if(token!=0) {
-	    fprintf(stderr, "tagged item image[%x]=%llx unrelocated\n", i, (long long)image[i]);
-	  }
-        }
+	}
       }
     }
+    free(targets);
+    if(ii==0)
+      image[0] = (Cell)image;
+    finish_code();
   }
-  free(targets);
-  finish_code();
-  ((ImageHeader*)(image))->base = (Address) image;
 }
 
 #ifndef DOUBLY_INDIRECT
@@ -918,7 +932,7 @@ static void check_prims(Label symbols1[])
 {
   int i;
 #ifndef NO_DYNAMIC
-  Label *symbols2, *symbols3, *ends1, *ends1j, *ends1jsorted, *goto_p;
+  Label *symbols2, *ends1, *ends1j, *ends1jsorted, *goto_p;
   int nends1j;
 #endif
 
@@ -938,17 +952,12 @@ static void check_prims(Label symbols1[])
   if (no_dynamic)
     return;
   symbols2=gforth_engine2(0 sr_call);
-#if NO_IP
-  symbols3=gforth_engine3(0 sr_call);
-#else
-  symbols3=symbols1;
-#endif
   ends1 = symbols1+i+1;
   ends1j =   ends1+i;
   goto_p = ends1j+i+1; /* goto_p[0]==before; ...[1]==after;*/
   nends1j = i+1;
   ends1jsorted = (Label *)alloca(nends1j*sizeof(Label));
-  memcpy(ends1jsorted,ends1j,nends1j*sizeof(Label));
+  memmove(ends1jsorted,ends1j,nends1j*sizeof(Label));
   qsort(ends1jsorted, nends1j, sizeof(Label), compare_labels);
 
   /* check whether the "goto *" is relocatable */
@@ -972,7 +981,6 @@ static void check_prims(Label symbols1[])
     int j=0;
     char *s1 = (char *)symbols1[i];
     char *s2 = (char *)symbols2[i];
-    char *s3 = (char *)symbols3[i];
     Label endlabel = bsearch_next(symbols1[i]+1,ends1jsorted,nends1j);
 
     pi->start = s1;
@@ -990,9 +998,9 @@ static void check_prims(Label symbols1[])
 	debugp(stderr, "S%d: op%d(S%d) = %d (%d);", sc->state_in, p, sc->state_out, i+1, pi->length);
     }
 #else
-    debugp(stderr, "%-15s %d-%d %4d %p %p %p len=%3ld rest=%2ld send=%1d",
+    debugp(stderr, "%-15s %d-%d %4d %p %p len=%3ld rest=%2ld send=%1d",
 	   prim_names[i], sc->state_in, sc->state_out,
-	   i, s1, s2, s3, (long)(pi->length), (long)(pi->restlength),
+	   i, s1, s2, (long)(pi->length), (long)(pi->restlength),
 	   pi->superend);
 #endif
     if (endlabel == NULL) {
@@ -1043,42 +1051,17 @@ static void check_prims(Label symbols1[])
 #endif
     };
     while (j<(pi->length+pi->restlength)) {
-      if (s1[j]==s3[j]) {
-	if (s1[j] != s2[j]) {
-	  pi->start = NULL; /* not relocatable */
+      if (s1[j] != s2[j]) {
+	pi->start = NULL; /* not relocatable */
 #ifndef BURG_FORMAT
-	  debugp(stderr,"\n   non_reloc: engine1!=engine2 offset %3d",j);
+	debugp(stderr,"\n   non_reloc: engine1!=engine2 offset %3d",j);
 #endif
-	  /* assert(j<prim_len); */
-	  relocs--;
-	  nonrelocs++;
-	  break;
-	}
-	j++;
-      } else {
-	struct immarg *ia=&pi->immargs[pi->nimmargs];
-
-	pi->nimmargs++;
-	ia->offset=j;
-	if ((~*(Cell *)&(s1[j]))==*(Cell *)&(s3[j])) {
-	  ia->rel=0;
-	  debugp(stderr,"\n   absolute immarg: offset %3d",j);
-	} else if ((&(s1[j]))+(*(Cell *)&(s1[j]))+4 ==
-		   symbols1[DOER_MAX+1]) {
-	  ia->rel=1;
-	  debugp(stderr,"\n   relative immarg: offset %3d",j);
-	} else {
-	  pi->start = NULL; /* not relocatable */
-#ifndef BURG_FORMAT
-	  debugp(stderr,"\n   non_reloc: engine1!=engine3 offset %3d",j);
-#endif
-	  /* assert(j<prim_len);*/
-	  relocs--;
-	  nonrelocs++;
-	  break;
-	}
-	j+=4;
+	/* assert(j<prim_len); */
+	relocs--;
+	nonrelocs++;
+	break;
       }
+      j++;
     }
     debugp(stderr,"\n");
   }
@@ -1110,7 +1093,7 @@ static void MAYBE_UNUSED align_code(void)
   UCell offset = ((UCell)code_here)&(alignment-1);
   UCell length = alignment-offset;
   if (length <= maxpadding) {
-    memcpy(code_here,nops+offset,length);
+    memmove(code_here,nops+offset,length);
     code_here += length;
   }
 #endif /* defined(CODE_PADDING) */
@@ -1124,10 +1107,10 @@ static void append_jump(void)
     PrimInfo *pi = &priminfos[last_jump];
     
     /* debugp(stderr, "Copy code %p<=%p+%x,%d\n", code_here, pi->start, pi->length, pi->restlength); */
-    memcpy(code_here, pi->start+pi->length, pi->restlength);
+    memmove(code_here, pi->start+pi->length, pi->restlength);
     code_here += pi->restlength;
     /* debugp(stderr, "Copy goto %p<=%p,%d\n", code_here, goto_start, goto_len); */
-    memcpy(code_here, goto_start, goto_len);
+    memmove(code_here, goto_start, goto_len);
     code_here += goto_len;
     align_code();
     last_jump=0;
@@ -1180,7 +1163,7 @@ static Address append_prim(Cell p)
   if(reserve_code_space(pi->length+pi->restlength+goto_len+CODE_ALIGNMENT-1))
     return NULL;
   /* debugp(stderr, "Copy code %p<=%p,%d\n", code_here, pi->start, pi->length); */
-  memcpy(code_here, pi->start, pi->length);
+  memmove(code_here, pi->start, pi->length);
   old_code_here = code_here;
   code_here += pi->length;
   return old_code_here;
@@ -1272,144 +1255,22 @@ Label decompile_code(Label _code)
 #endif /* !defined(NO_DYNAMIC) */
 }
 
-#ifdef NO_IP
-int nbranchinfos=0;
-
-struct branchinfo {
-  Label **targetpp; /* **(bi->targetpp) is the target */
-  Cell *addressptr; /* store the target here */
-} branchinfos[100000];
-
-int ndoesexecinfos=0;
-struct doesexecinfo {
-  int branchinfo; /* fix the targetptr of branchinfos[...->branchinfo] */
-  Label *targetp; /*target for branch (because this is not in threaded code)*/
-  Cell *xt; /* cfa of word whose does-code needs calling */
-} doesexecinfos[10000];
-
-static void set_rel_target(Cell *source, Label target)
-{
-  *source = ((Cell)target)-(((Cell)source)+4);
-}
-
-static void register_branchinfo(Label source, Cell *targetpp)
-{
-  struct branchinfo *bi = &(branchinfos[nbranchinfos]);
-  bi->targetpp = (Label **)targetpp;
-  bi->addressptr = (Cell *)source;
-  nbranchinfos++;
-}
-
-static Address compile_prim1arg(PrimNum p, Cell **argp)
-{
-  Address old_code_here=append_prim(p);
-  if(old_code_here==NULL)
-    return NULL;
-
-  assert(vm_prims[p]==priminfos[p].start);
-  *argp = (Cell*)(old_code_here+priminfos[p].immargs[0].offset);
-  return old_code_here;
-}
-
-static Address compile_call2(Cell *targetpp, Cell **next_code_targetp)
-{
-  PrimInfo *pi = &priminfos[N_call2];
-  Address old_code_here = append_prim(N_call2);
-  if(old_code_here==NULL)
-    return NULL;
-
-  *next_code_targetp = (Cell *)(old_code_here + pi->immargs[0].offset);
-  register_branchinfo(old_code_here + pi->immargs[1].offset, targetpp);
-  return old_code_here;
-}
-#endif
-
 void finish_code(void)
 {
-#ifdef NO_IP
-  Cell i;
+  compile_prim1(NULL);
+  flush_to_here();
+}
 
+void finish_code_barrier(void)
+{
   compile_prim1(NULL);
-  for (i=0; i<ndoesexecinfos; i++) {
-    struct doesexecinfo *dei = &doesexecinfos[i];
-    dei->targetp = (Label *)DOES_CODE1((dei->xt));
-    branchinfos[dei->branchinfo].targetpp = &(dei->targetp);
-  }
-  ndoesexecinfos = 0;
-  for (i=0; i<nbranchinfos; i++) {
-    struct branchinfo *bi=&branchinfos[i];
-    set_rel_target(bi->addressptr, **(bi->targetpp));
-  }
-  nbranchinfos = 0;
-#else
-  compile_prim1(NULL);
+#ifndef NO_DYNAMIC
+  append_jump();
 #endif
   flush_to_here();
 }
 
 #if !(defined(DOUBLY_INDIRECT) || defined(INDIRECT_THREADED))
-#ifdef NO_IP
-static Cell compile_prim_dyn(PrimNum p, Cell *tcp)
-     /* compile prim #p dynamically (mod flags etc.) and return start
-	address of generated code for putting it into the threaded
-	code. This function is only called if all the associated
-	inline arguments of p are already in place (at tcp[1] etc.) */
-{
-  PrimInfo *pi=&priminfos[p];
-  Cell *next_code_target=NULL;
-  Address codeaddr;
-  Address primstart;
-  
-  assert(p<npriminfos);
-  if (p==N_execute || p==N_perform || p==N_lit_perform) {
-    if((codeaddr = compile_prim1arg(N_set_next_code, &next_code_target)) == NULL)
-      return 0;
-    primstart = append_prim(p);
-    goto other_prim;
-  } else if (p==N_call) {
-    if((codeaddr = compile_call2(tcp+1, &next_code_target)) == NULL)
-      return 0;
-  } else if (p==N_does_exec) {
-    struct doesexecinfo *dei = &doesexecinfos[ndoesexecinfos++];
-    Cell *arg;
-    if((codeaddr = compile_prim1arg(N_lit,&arg)) == NULL)
-      return 0;
-    *arg = (Cell)PFA(tcp[1]);
-    /* we cannot determine the callee now (last_start[1] may be a
-       forward reference), so just register an arbitrary target, and
-       register in dei that we need to fix this before resolving
-       branches */
-    dei->branchinfo = nbranchinfos;
-    dei->xt = (Cell *)(tcp[1]);
-    if(compile_call2(0, &next_code_target)==NULL)
-      return 0;
-  } else if (!is_relocatable(p)) {
-    Cell *branch_target;
-    if((codeaddr = compile_prim1arg(N_set_next_code, &next_code_target)) == NULL)
-      return 0;
-    if(compile_prim1arg(N_branch,&branch_target)==NULL)
-      return 0;
-    set_rel_target(branch_target,vm_prims[p]);
-  } else {
-    unsigned j;
-
-    codeaddr = primstart = append_prim(p);
-  other_prim:
-    for (j=0; j<pi->nimmargs; j++) {
-      struct immarg *ia = &(pi->immargs[j]);
-      Cell *argp = tcp + pi->nimmargs - j;
-      Cell argval = *argp; /* !! specific to prims */
-      if (ia->rel) { /* !! assumption: relative refs are branches */
-	register_branchinfo(primstart + ia->offset, argp);
-      } else /* plain argument */
-	*(Cell *)(primstart + ia->offset) = argval;
-    }
-  }
-  if (next_code_target!=NULL)
-    *next_code_target = (Cell)code_here;
-  return (Cell)codeaddr;
-}
-#else /* !defined(NO_IP) */
 static Cell compile_prim_dyn(PrimNum p, Cell *tcp)
      /* compile prim #p dynamically (mod flags etc.) and return start
         address of generated code for putting it into the threaded code */
@@ -1433,7 +1294,6 @@ static Cell compile_prim_dyn(PrimNum p, Cell *tcp)
   return (Cell)old_code_here;
 #endif  /* !defined(NO_DYNAMIC) */
 }
-#endif /* !defined(NO_IP) */
 #endif
 
 #ifndef NO_DYNAMIC
@@ -1993,7 +1853,7 @@ static FILE *checkimage(char *path, int len, char *imagename)
 #endif
     ;
 
-  memcpy(fullfilename, path, dirlen);
+  memmove(fullfilename, path, dirlen);
   if (dirlen && fullfilename[dirlen-1]!=DIRSEP)
     fullfilename[dirlen++]=DIRSEP;
   strcpy(fullfilename+dirlen,imagename);
@@ -2009,7 +1869,7 @@ static FILE *checkimage(char *path, int len, char *imagename)
       return NULL;
     }
     preamblesize+=8;
-  } while(memcmp(magic,"Gforth5",7));
+  } while(memcmp(magic,"Gforth6",7));
   if (debug) {
     fprintf(stderr,"Magic found: %*s ", 6, magic);
     print_sizes(magic[7]);
@@ -2072,6 +1932,11 @@ ImageHeader* gforth_loader(char* imagename, char* path)
 /* returns the address of the image proper (after the preamble) */
 {
   ImageHeader header;
+  SectionHeader section;
+  Address sections[0x100];  /* base address of all sections */
+  Char* reloc_bits[0x100];  /* reloc bits of all images     */
+  UCell sizes[0x100];       /* section sizes */
+  Cell bases[0x100];        /* section bases */
   Address image;
   Address imp; /* image+preamble */
   Cell data_offset = offset_image ? 56*sizeof(Cell) : 0;
@@ -2101,6 +1966,8 @@ ImageHeader* gforth_loader(char* imagename, char* path)
   }
 
   set_stack_sizes(&header);
+  bzero(sections, sizeof(sections));
+  bzero(reloc_bits, sizeof(reloc_bits));
   
 #if HAVE_GETPAGESIZE
   pagesize=getpagesize(); /* Linux/GNU libc offers this */
@@ -2111,36 +1978,62 @@ ImageHeader* gforth_loader(char* imagename, char* path)
 #endif
   debugp(stderr,"pagesize=%ld\n",(unsigned long) pagesize);
 
-  image = dict_alloc_read(imagefile, preamblesize+header.image_size,
+  sizes[0]=header.image_dp-header.base;
+  bases[0]=(Cell)header.base;
+
+  image = dict_alloc_read(imagefile, preamblesize+sizes[0],
 			  dictsize, data_offset);
   if(image==NULL) return NULL;
-  imp=image+preamblesize;
+
+  sections[0]=imp=image+preamblesize;
 
   set_stack_sizes((ImageHeader*)imp);
 
   if (clear_dictionary)
-    memset(imp+header.image_size, 0, dictsize-header.image_size-preamblesize);
-  if(header.base==0 || header.base  == (Address)0x100) {
-    Cell reloc_size=((header.image_size-1)/sizeof(Cell))/8+1;
-    Char reloc_bits[reloc_size];
-    fseek(imagefile, preamblesize+header.image_size, SEEK_SET);
-    if(reloc_size != fread(reloc_bits, 1, reloc_size, imagefile)) {
-      fprintf(stderr, "Image reloc bits read terminated early\n");
+    memset(imp+sizes[0], 0, dictsize-sizes[0]-preamblesize);
+  
+  fseek(imagefile, preamblesize+sizes[0], SEEK_SET);
+    
+  int i;
+  for(i=0; i<0xFE; ) {
+    Cell reloc_size=((sizes[i]-1)/sizeof(Cell))/8+1;
+    if(bases[i]==0 || bases[i] == 0x100) {
+      reloc_bits[i]=malloc(reloc_size);
+      
+      if(reloc_size != fread(reloc_bits[i], 1, reloc_size, imagefile)) {
+	fprintf(stderr, "Image reloc bits read terminated early\n");
+	break;
+      }
+    } else if(bases[i]!=(Cell)sections[i]) {
+      fprintf(stderr,"%s: Cannot load nonrelocatable image (compiled for address %p) at address %p\n",
+	      progname, (Address)bases[i], sections[i]);
+      return NULL;
     }
-    gforth_relocate((Cell *)imp, reloc_bits, header.image_size, (Cell)header.base, vm_prims);
+    
+    if(8 != fread(magic, 1, 8, imagefile)) break;
+    
+    if(memcmp(magic, "Section.", 8)) break;
+    
+    if(sizeof(SectionHeader) !=
+       fread(&section, 1, sizeof(SectionHeader), imagefile)) break;
+    
+    i++;
+    
+    bases[i] = INSECTION(section.base);
+    sizes[i] = section.dp-section.base;
+    sections[i] = alloc_mmap_guard(section.end);
+    fseek(imagefile, -sizeof(SectionHeader), SEEK_CUR);
+    debugp(stderr, "section base=%p, dp=%p, end=%p\n", section.base, section.dp, section.end);
+    if(fread(sections[i], 1, sizes[i], imagefile) != sizes[i]) break;
+  }
+  gforth_relocate(sections, reloc_bits, sizes, bases, vm_prims);
 #if 0
-    { /* let's see what the relocator did */
-      FILE *snapshot=fopen("snapshot.fi","wb");
-      fwrite(image,1,imagesize,snapshot);
-      fclose(snapshot);
-    }
+  { /* let's see what the relocator did */
+    FILE *snapshot=fopen("snapshot.fi","wb");
+    fwrite(image,1,imagesize,snapshot);
+    fclose(snapshot);
+  }
 #endif
-  }
-  else if(header.base!=imp) {
-    fprintf(stderr,"%s: Cannot load nonrelocatable image (compiled for address %p) at address %p\n",
-	    progname, header.base, imp);
-    return NULL;
-  }
   if (header.checksum==0)
     ((ImageHeader *)imp)->checksum=check_sum;
   else if (header.checksum != check_sum) {
@@ -2154,6 +2047,10 @@ ImageHeader* gforth_loader(char* imagename, char* path)
 #endif
   fclose(imagefile);
 
+  for(i=0; i<0x100; i++) {
+    if(reloc_bits[i]!=NULL)
+      free(reloc_bits[i]);
+  }
   /* unnecessary, except maybe for CODE words */
   /* FLUSH_ICACHE(imp, header.image_size);*/
 
