@@ -1,5 +1,6 @@
 /* a simple pow10 implementation
 
+  Authors: Anton Ertl, Bernd Paysan
   Copyright (C) 1995,1998,2000,2003,2007,2017 Free Software Foundation, Inc.
 
   This file is part of Gforth.
@@ -19,6 +20,7 @@
 */
 #include "config.h"
 #include <math.h>
+#include "symver.h"
 
 #ifndef M_LN10
 #define M_LN10      2.30258509299404568402
