@@ -1,5 +1,6 @@
 /* Gforth support functions
 
+  Authors: Bernd Paysan, Anton Ertl
   Copyright (C) 1995,1996,1997,1998,2000,2003,2004,2006,2007,2008,2009,2010,2011,2012,2013,2014,2015,2016,2017,2018 Free Software Foundation, Inc.
 
   This file is part of Gforth.
@@ -21,6 +22,7 @@
 #include "config.h"
 #include "forth.h"
 #include "io.h"
+#include "symver.h"
 #include <stdlib.h>
 #include <string.h>
 #include <sys/time.h>
@@ -113,7 +115,7 @@ char *cstr(Char *from, UCell size)
    the C-string lives until free */
 {
   char * string = malloc_l(size+1);
-  memcpy(string,from,size);
+  memmove(string,from,size);
   string[size]='\0';
   return string;
 }
@@ -169,7 +171,7 @@ char *tilde_cstr(Char *from, UCell size)
       return cstr(from+3, size<3?0:size-3);
     {
       char user[i];
-      memcpy(user,from+1,i-1);
+      memmove(user,from+1,i-1);
       user[i-1]='\0';
       user_entry=getpwnam(user);
     }
@@ -184,8 +186,8 @@ char *tilde_cstr(Char *from, UCell size)
     s1_len--;
   {
     char path[s1_len+s2_len];
-    memcpy(path,s1,s1_len);
-    memcpy(path+s1_len,s2,s2_len);
+    memmove(path,s1,s1_len);
+    memmove(path+s1_len,s2,s2_len);
     if(allocs1) free_l(s1);
     return cstr((Char *)path,s1_len+s2_len);
   }
@@ -388,7 +390,7 @@ void hashkey2(Char* c_addr, UCell u, uint64_t upmask, hash128 *h)
   Char* endp = c_addr+(u&-sizeof(uint64_t)), *endp1=c_addr+u-sizeof(uint64_t);
   // read all full words
   for(; c_addr<endp; c_addr+=sizeof(uint64_t)) {
-    memcpy(&mixin, c_addr, sizeof(uint64_t));
+    memmove(&mixin, c_addr, sizeof(uint64_t));
     // printf("+%lx\n", mixin);
     mixin |= upmask & ~(mixin >> 2); // case insensitive trick
     a ^= mixin;
@@ -404,7 +406,7 @@ void hashkey2(Char* c_addr, UCell u, uint64_t upmask, hash128 *h)
       endp = endp1;
       lastshift= shift;
     }
-    memcpy(&mixin, endp, sizeof(uint64_t));
+    memmove(&mixin, endp, sizeof(uint64_t));
     // strip off parts read after the string end
     // and mix in length of remaining fragment
 #ifdef WORDS_BIGENDIAN
@@ -446,14 +448,14 @@ UCell hashkey2a(Char *s, UCell n)
     if (((((UCell)(s+n-1))^((UCell)(s+w-1)))&(-pagesize)) != 0) {
       /* cell access would cross page boundary, but byte access wouldn't */
       /* so cell access to s might incur a SIGSEGV */
-      memcpy(&h,(char *)(s+n-w),w);
+      memmove(&h,(char *)(s+n-w),w);
 #ifdef WORDS_BIGENDIAN
       h = h & ((~(UCell)0) >> erase);
 #else
       h = h>>erase;
 #endif
     } else {
-      memcpy(&h,s,w);
+      memmove(&h,s,w);
 #ifdef WORDS_BIGENDIAN
       h = h>>erase;
 #else
@@ -468,14 +470,14 @@ UCell hashkey2a(Char *s, UCell n)
     Char *p = s;
     h = seed;
     do {
-      memcpy(&h1,p,w);
+      memmove(&h1,p,w);
       h1 |= upmask & ~(h1 >> 2); // case insensitive trick
       h = (h^h1)*k0;
       h = rotl(h,rot1);
       p += w;
     } while (p<s+n-w);
     p = s+n-w;
-    memcpy(&h1,p,w);
+    memmove(&h1,p,w);
     h1 |= upmask & ~(h1 >> 2); // case insensitive trick
     h = (h^h1)*k0;
     h = rotl(h,rot1);
@@ -723,8 +725,8 @@ int gforth_system(Char *c_addr, UCell u)
   old_tp=terminal_prepped;
   deprep_terminal();
 #endif
-  memcpy(buffer,prefix,prefixlen);
-  memcpy(buffer+prefixlen,c_addr,u);
+  memmove(buffer,prefix,prefixlen);
+  memmove(buffer+prefixlen,c_addr,u);
   buffer[prefixlen+u]='\0';
   retval=system(buffer); /* ~ expansion on first part of string? */
 #ifndef MSDOS
