@@ -1,5 +1,6 @@
 ;;; gforth.el --- major mode for editing (G)Forth sources
 
+;; Authors: Bernd Paysan, Anton Ertl, David Kühling
 ;; Copyright (C) 1995,1996,1997,1998,2000,2001,2003,2004,2007,2008,2010,2011,2012,2013,2014,2015,2016,2017,2018 Free Software Foundation, Inc.
 
 ;; This file is part of Gforth.
@@ -237,6 +238,10 @@ PARSED-TYPE specifies what kind of text is parsed. It should be on of 'name',
 	(("]" "]l") definition-starter (font-lock-keyword-face . 1))
 	((":") definition-starter (font-lock-keyword-face . 1)
 	 "[ \t\n]" t name (font-lock-function-name-face . 3))
+	(("to:") definition-starter (font-lock-keyword-face . 1)
+	 "[ \t\n]" t name (font-lock-function-name-face . 3))
+	(("defer@:") definition-starter (font-lock-keyword-face . 1)
+	 "[ \t\n]" t name (font-lock-function-name-face . 3))
 	(("event:") definition-starter (font-lock-keyword-face . 1)
 	 "[ \t\n]" t name (font-lock-function-name-face . 3))
 	(("immediate" "compile-only" "restrict")
@@ -259,7 +264,7 @@ PARSED-TYPE specifies what kind of text is parsed. It should be on of 'name',
 	(("char") non-immediate (font-lock-keyword-face . 1)
 	 "[ \t\n]" t string (font-lock-string-face . 1))
 	("'.'?" non-immediate (font-lock-string-face . 1))
-	(("s\"" "c\"" "s\\\"") immediate (font-lock-string-face . 1)
+	(("s\"" "c\"" "l\"" "x\"" "s\\\"") immediate (font-lock-string-face . 1)
 	 "[\"\n]" nil string (font-lock-string-face . 1))
 	((".\"" ".\\\"") compile-only (font-lock-string-face . 1)
 	 "[\"\n]" nil string (font-lock-string-face . 1))
@@ -278,7 +283,9 @@ PARSED-TYPE specifies what kind of text is parsed. It should be on of 'name',
 	  
 	(("[if]" "[?do]" "[do]" "[for]" "[begin]" 
 	  "[endif]" "[then]" "[loop]" "[+loop]" "[next]" "[until]" "[repeat]"
-	  "[again]" "[while]" "[else]" "[:" "[{:" ";]" "nope")
+	  "[again]" "[while]" "[else]" "[:"
+	  "[n:l" "[n:h" "[n:d" "[d:l" "[d:h" "[d:d" "[f:l" "[f:h" "[f:d"
+	  "[{:" ";]" "nope")
 	 immediate (font-lock-keyword-face . 2))
 	(("[ifdef]" "[ifundef]" "[defined]" "[undefined]") immediate (font-lock-keyword-face . 2)
 	 "[ \t\n]" t name (font-lock-function-name-face . 3))
@@ -320,13 +327,17 @@ PARSED-TYPE specifies what kind of text is parsed. It should be on of 'name',
 	  "+field" "value:" "cvalue:" "scvalue:" "wvalue:" "swvalue:"
 	  "lvalue:" "slvalue:" "2value:" "fvalue:" "sfvalue:" "dfvalue:"
 	  "$value:" "defer:" "value[]:" "$value[]:"
-	  "wrap+value:")
+	  "wrap+value:" "method" "umethod")
 	 non-immediate (font-lock-type-face . 2)
 	 "[ \t\n]" t name (font-lock-variable-name-face . 3))
 	("\\S-+%" non-immediate (font-lock-type-face . 2))
 	(("defer" "alias" "create-interpret/compile:") 
 	 non-immediate (font-lock-type-face . 1)
 	 "[ \t\n]" t name (font-lock-function-name-face . 3))
+	(("synonym") 
+	 non-immediate (font-lock-type-face . 1)
+	 "[ \t\n]" t name (font-lock-function-name-face . 3)
+	 "[ \t\n]" t name (font-lock-function-name-face . 2))
 	(("end-struct") non-immediate (font-lock-keyword-face . 2)
 	 "[ \t\n]" t name (font-lock-type-face . 3))
 	(("struct" "end-c-library" "c-library-name" "end-structure") 
@@ -384,7 +395,7 @@ PARSED-TYPE specifies what kind of text is parsed. It should be on of 'name',
      "[ \t\n]" t name (font-lock-function-name-face . 3))
     (("inst-var" "inst-value") non-immediate (font-lock-type-face . 2)
      "[ \t\n]" t name (font-lock-variable-name-face . 3))
-    (("method" "selector")
+    (("method" "selector" "umethod")
      non-immediate (font-lock-type-face . 1)
      "[ \t\n]" t name (font-lock-function-name-face . 3))
     (("end-class" "end-interface")
@@ -466,9 +477,10 @@ INDENT1 and INDENT2 are indentation specifications of the form
 (setq forth-indent-words
       '((("if" "begin" "do" "?do" "+do" "-do" "u+do"
 	  "u-do" "?dup-if" "?dup-0=-if" "case" "of" "?of" "try" "iferror"
-	  "[if]" "[ifdef]" "[ifundef]" "[begin]" "[for]" "[do]" "[?do]" "[:" "[{:")
+	  "[if]" "[ifdef]" "[ifundef]" "[begin]" "[for]" "[do]" "[?do]" "[:"
+	  "[n:l" "[n:h" "[n:d" "[d:l" "[d:h" "[d:d" "[f:l" "[f:h" "[f:d" "[{:")
 	 (0 . 2) (0 . 2))
-	((":" ":noname" "code" "abi-code" "struct" "m:" ":m" "class" 
+	((":" ":noname" "code" "abi-code" "struct" "m:" ":m" "class" "uclass" 
 	  "interface" "c-library" "c-library-name" "comp:" "opt:" "post:"
 	  "begin-structure" "extend-structure" "event:" "to-opt:" "defer@-opt:" "to:" "defer@:")
 	 (0 . 2) (0 . 2) non-immediate)
@@ -788,7 +800,7 @@ End:\" construct).")
   '("VARIABLE" "CONSTANT" "2VARIABLE" "2CONSTANT" "FVARIABLE" "FCONSTANT"
    "USER" "VALUE" "2Value" "field" "end-struct" "VOCABULARY" "CREATE" ":" "CODE"
    "DEFER" "ALIAS" "interpret/compile:" "debug:" "field:" "2field:" "ffield:"
-   "sffield:" "dffield:" "uvar" "uvalue" "voctable")
+   "sffield:" "dffield:" "uvar" "uvalue" "voctable" "method" "umethod")
   "List of words, that define the following word.
 Used for imenu index generation.")
 
@@ -1213,6 +1225,7 @@ exceeds 64 characters."
   ;(setq comment-end " )")
   (make-local-variable 'comment-column)
   (setq comment-column 40)
+  (setq compilation-error-screen-columns nil) ;; for WHEREG
   (make-local-variable 'comment-start-skip)
   (setq comment-start-skip "\\\\ ")
   (make-local-variable 'comment-indent-function)
@@ -1377,13 +1390,11 @@ bell during block file read/write operations."
 	   (forth-change-function (point-min) (point-max) nil t))))
 
 (defun forth-fill-paragraph () 
-  "Fill comments (starting with '\'; do not fill code (block style
-programmers who tend to fill code won't use emacs anyway:-)."
-  ; Currently only comments at the start of the line are filled.
-  ; Something like lisp-fill-paragraph may be better.  We cannot use
-  ; fill-paragraph, because it removes the \ from the first comment
-  ; line. Therefore we have to look for the first line of the comment
-  ; and use fill-region.
+  "Fill comments starting with '\\' which start a line; do not fill code."
+  ;; Something like lisp-fill-paragraph may be better.  We cannot use
+  ;; fill-paragraph, because it removes the \ from the first comment
+  ;; line. Therefore we have to look for the first line of the comment
+  ;; and use fill-region.
   (interactive)
   (save-excursion
     (beginning-of-line)
@@ -1396,8 +1407,10 @@ programmers who tend to fill code won't use emacs anyway:-)."
 	  (to (save-excursion (forward-paragraph) (point))))
       (if (looking-at "[ \t]*\\\\g?[ \t]+")
 	  (progn (goto-char (match-end 0))
-		 (set-fill-prefix)
-		 (fill-region from to nil))))))
+		 (let ((fill-prefix-save fill-prefix))
+		   (set-fill-prefix)
+		   (fill-region from to nil)
+		   (setq fill-prefix fill-prefix-save)))))))
 
 (defun forth-comment-indent ()
   (save-excursion
